from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

import gurobipy as gp
import polars as pl

from xplor.exprs import VarExpr

if TYPE_CHECKING:
    from collections.abc import Sequence


def to_mvar_or_mlinexpr(s: pl.Series) -> gp.MVar | gp.MLinExpr | Any:
    """Convert a Polars Series to Gurobi MVar, MLinExpr, or NumPy array.

    Parameters
    ----------
    s : pl.Series
        Series containing Gurobi variables, linear expressions, or numeric values.

    Returns
    -------
    gp.MVar | gp.MLinExpr | np.ndarray
        Vectorized Gurobi array or NumPy array.

    """
    first: Any = s.first(ignore_nulls=True)
    if isinstance(first, gp.Var):
        return gp.MVar(s)  # ty:ignore[too-many-positional-arguments]
    if isinstance(first, gp.LinExpr):
        return gp.MLinExpr(s.to_numpy())  # ty:ignore[too-many-positional-arguments]
    return s.to_numpy()


class GurobiVarExpr(VarExpr):
    """Gurobi-specific VarExpr with optimized vectorized operations.

    Extends VarExpr with Gurobi-specific optimizations using MVar/MLinExpr
    for efficient vectorized evaluation.

    Methods
    -------
        sum(): Vectorized sum using MVar/MLinExpr.
        any(): Creates a Gurobi OR constraint across elements in each group.
        all(): Creates a Gurobi AND constraint across elements in each group.
        abs(): Applies Gurobi's absolute value function.
        max(): Returns the maximum of elements in each group.
        min(): Returns the minimum of elements in each group.

    """

    def _to_expr(self) -> pl.Expr:
        if not self._nodes:
            return self._expr
        expr_repr, exprs = self.parse()

        def vectorized_eval(series: Sequence[pl.Series], **kwargs: Any) -> pl.Series:
            """Evaluate expression using vectorized NumPy operations."""
            gp_obj = tuple(map(to_mvar_or_mlinexpr, series))
            result = expr_repr.evaluate(gp_obj)

            # Unfortunalty `MLinExpr` need to be disptach into list of LinExpr from the python side.
            if isinstance(result, gp.MLinExpr):
                result = [r.item() for r in result]
            return pl.Series(result, dtype=pl.Object)

        return pl.map_batches(
            exprs,
            vectorized_eval,
            return_dtype=pl.Object,
        )

    def sum(
        self,
    ) -> Self:
        """Vectorized sum using Gurobi MVar/MLinExpr for optimal performance.

        For expressions with nodes (e.g., var.x + var.y)
        to evaluate vectorized, then calls .sum() on the result. This is
        2-3x faster than element-wise summation.

        Optimized pattern: (var * coeff).sum() uses gp.LinExpr(coeffs, vars)
        directly, which is ~10x faster than element-wise multiplication + sum.

        Examples
        --------
        >>> df.group_by('group').agg(xmodel.var("x").sum())
        >>> df.select(total=(var.x + var.y).sum())  # Vectorized!
        >>> df.select(total=(var.x * pl.col("cost")).sum())  # Optimized LinExpr!

        """
        name = str(self) if self.meta.is_column() else f"({self})"
        expr_repr, exprs = self.parse()
        # Optimization: Detect (var * coeff).sum() pattern and use gp.LinExpr directly
        # This is ~10x faster than element-wise multiplication followed by sum
        if len(self._nodes) == 1 and self._nodes[0].operator in ("__mul__", "__rmul__"):

            def gurobi_vectorized_sum(series: Sequence[pl.Series]) -> Any:
                # If coeff are Object, fast path can't be taken
                if (len(series) >= 2 and series[1].dtype is not pl.Object) or (
                    isinstance(self._nodes[0].operand, float)
                ):
                    operand = (
                        series[1].to_list()
                        if len(series) > 1
                        else len(series[0]) * [self._nodes[0].operand]
                    )
                    result = gp.LinExpr(operand, series[0].to_list())  # type: ignore

                else:
                    gp_obj = tuple(map(to_mvar_or_mlinexpr, series))
                    result = expr_repr.evaluate(gp_obj)
                return pl.Series([result], dtype=pl.Object)
        else:
            # Standard vectorized sum for other expression patterns
            def gurobi_vectorized_sum(series: Sequence[pl.Series]) -> Any:
                gp_obj = tuple(map(to_mvar_or_mlinexpr, series))
                result = expr_repr.evaluate(gp_obj)
                return pl.Series([result.sum()], dtype=pl.Object)

        _, exprs = self.parse()
        return self.__class__(
            pl.map_batches(
                exprs,
                gurobi_vectorized_sum,
                return_dtype=pl.Object,
                returns_scalar=True,
            ),
            name=name + ".sum()",
        )

    def any(self) -> Self:  # ty:ignore[invalid-method-override]
        """Create a Gurobi OR constraint from elements in each group.

        Parameters
        ----------
        expr : pl.Expr | str
            Column name or polars expression containing Gurobi variables or expressions

        Examples
        --------
        >>> df.group_by('group').agg(xmodel.var("x").any())

        """
        return self.__class__(
            self.map_batches(
                lambda d: gp.or_(d.to_list()), return_dtype=pl.Object, returns_scalar=True
            ),
            name=f"{self}.any()",
        )

    def abs(self) -> Self:
        """Apply Gurobi's absolute value function to elements in each group.

        Parameters
        ----------
        expr : pl.Expr | str
            Column name or polars expression containing Gurobi variables or expressions

        Examples
        --------
        >>> df.with_columns(xmodel.var("x").abs())

        """
        return self.__class__(
            self.map_elements(lambda d: gp.abs_(d), return_dtype=pl.Object),
            name=f"{self}.abs()",
        )


class _ProxyGurobiVarExpr:
    def __call__(self, name: str, /, *more_names: str) -> GurobiVarExpr:
        return GurobiVarExpr(pl.col(name, *more_names))

    def __getattr__(self, name: str) -> GurobiVarExpr:
        return GurobiVarExpr(pl.col(name))



import polars as pl

from xplor.typing import VariableType


def cast_to_dtypes(series: pl.Series, variable_type: VariableType | str) -> pl.Series:
    """Cast a series to the corresponding data type base on its variable_type.

    Note: Accepts str to handle internal markers like "_INTERVAL_CP" for CPLEX CP backend.
    """
    if variable_type == "INTEGER":
        return series.cast(pl.Int32)
    elif variable_type == "BINARY":
        return series.cast(pl.Int8).cast(pl.Boolean)
    elif variable_type == "_INTERVAL_CP":
        # Interval variables (CPLEX CP) are already returned as structs, no casting needed
        return series
    else:
        return series.cast(pl.Float64)



"""Interval variable expressions for CPLEX CP backend."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

import polars as pl
from docplex.cp.modeler import (
    abs as cp_abs,
)
from docplex.cp.modeler import (
    ceil,
    end_at_end,
    end_before_end,
    end_before_start,
    end_of,
    exponent,
    floor,
    length_of,
    log,
    no_overlap,
    power,
    presence_of,
    sgn,
    size_of,
    square,
    start_at_end,
    start_at_start,
    start_before_end,
    start_before_start,
    start_of,
    trunc,
)
from docplex.cp.modeler import (
    max as cp_max,
)
from docplex.cp.modeler import (
    min as cp_min,
)
from docplex.cp.modeler import (
    round as cp_round,
)
from docplex.cp.modeler import (
    sum as cp_sum,
)

from xplor._utils import series_to_df
from xplor.exprs import ConstrExpr, VarExpr

if TYPE_CHECKING:
    from docplex.cp.model import CpoModel


class IntervalVarExpr(VarExpr):
    """Expression wrapper for CPLEX CP interval variables.

    Provides methods for creating CP constraints on interval variables such as
    precedence, synchronization, and resource constraints.

    Examples
    --------
    >>> from xplor.cplex_cp import XplorCplexCP  # doctest: +SKIP
    >>> xmodel = XplorCplexCP()  # doctest: +SKIP
    >>> # Create precedence constraint
    >>> xmodel.add_constr(xmodel.var.task1.end_before_start(xmodel.var.task2))  # doctest: +SKIP

    """

    _model: CpoModel

    def _create_binary_operation(
        self,
        other: VarExpr | pl.Expr | float,
        operation_func: Callable[[Any, Any], Any],
    ) -> VarExpr:
        """Create binary mathematical operation on interval expressions.

        Parameters
        ----------
        other : VarExpr | pl.Expr | float
            The second operand
        operation_func : Callable
            The docplex.cp.modeler function to apply (e.g., power, min, max)

        Returns
        -------
        IntervalVarExpr
            Expression representing the operation result

        """
        if not isinstance(other, pl.Expr):
            other = pl.lit(other)

        def func(df: pl.DataFrame) -> pl.Series:
            return pl.Series(
                [operation_func(row[0], row[1]) for row in df.rows()],
                dtype=pl.Object,
            )

        return IntervalVarExpr(
            pl.map_batches(
                [self, other],
                lambda series: func(series_to_df(series)),
                return_dtype=pl.Object,
            )
        )

    def _create_pairwise_constraint(
        self,
        other: IntervalVarExpr | pl.Expr,
        delay: pl.Expr | float,
        constraint_func: Callable[[Any, Any, float], Any],
    ) -> ConstrExpr:
        """Create pairwise interval constraint using docplex.cp.modeler function.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The other interval variable
        delay : pl.Expr | float
            Delay parameter
        constraint_func : Callable
            The docplex.cp.modeler constraint function to apply

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        if not isinstance(delay, pl.Expr):
            delay = pl.lit(delay)

        def func(df: pl.DataFrame) -> pl.Series:
            return pl.Series(
                [constraint_func(row[0], row[1], row[2]) for row in df.rows()],
                dtype=pl.Object,
            )

        return ConstrExpr(
            pl.map_batches(
                [self, other, delay],
                lambda series: func(series_to_df(series)),
                return_dtype=pl.Object,
            )
        )

    def end_before_start(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Create precedence constraint: this interval must end before other starts.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval that should start after this one ends
        delay : int | float | pl.Expr, default 0
            Minimum delay between end of this and start of other

        Returns
        -------
        ConstrExpr
            Constraint expression

        Examples
        --------
        >>> xmodel.add_constrs(df, precedence=var.task1.end_before_start(var.task2, delay=2))  # doctest: +SKIP

        """
        return self._create_pairwise_constraint(other, delay, end_before_start)

    def start_at_start(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Synchronize interval starts with optional delay.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval to synchronize with
        delay : int | float | pl.Expr, default 0
            Delay between starts

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, start_at_start)

    def end_at_end(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Synchronize interval ends with optional delay.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval to synchronize with
        delay : int | float | pl.Expr, default 0
            Delay between ends

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, end_at_end)

    def start_at_end(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Constrain this interval to start when other ends.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval that should end when this starts
        delay : int | float | pl.Expr, default 0
            Delay between end of other and start of this

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, start_at_end)

    def start_before_end(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Constrain this interval to start before other ends.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval that should end after this starts
        delay : int | float | pl.Expr, default 0
            Minimum delay

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, start_before_end)

    def start_before_start(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Constrain this interval to start before other starts.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval that should start after this starts
        delay : int | float | pl.Expr, default 0
            Minimum delay

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, start_before_start)

    def end_before_end(
        self, other: IntervalVarExpr | pl.Expr, delay: pl.Expr | float = 0
    ) -> ConstrExpr:
        """Constrain this interval to end before other ends.

        Parameters
        ----------
        other : IntervalVarExpr | pl.Expr
            The interval that should end after this ends
        delay : int | float | pl.Expr, default 0
            Minimum delay

        Returns
        -------
        ConstrExpr
            Constraint expression

        """
        return self._create_pairwise_constraint(other, delay, end_before_end)

    def no_overlap(self) -> ConstrExpr:
        """Create no-overlap constraint for intervals in this column.

        Ensures that intervals do not overlap in time (disjunctive constraint).
        This creates a single constraint that applies to all intervals in the column.

        Returns
        -------
        VarExpr
            Constraint expression containing the no_overlap constraint

        Examples
        --------
        >>> xmodel.add_constrs(df, machine1_no_overlap=var.task_iv.no_overlap())  # doctest: +SKIP

        """

        def func(series: pl.Series) -> pl.Series:
            # Get all interval variables from the column
            intervals = series.to_list()
            # Create a single no_overlap constraint
            constraint = no_overlap(intervals)
            # Return as a single-element series
            return pl.Series([constraint], dtype=pl.Object)

        return ConstrExpr(self.map_batches(func, return_dtype=pl.Object))

    def start_of(self) -> VarExpr:
        """Get the start time of interval variables.

        Returns
        -------
        VarExpr
            Expression representing the start time of each interval

        Examples
        --------
        >>> xmodel.add_constrs(df, start_constraint=var.task_iv.start_of() >= 10)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(start_of, return_dtype=pl.Object))

    def end_of(self) -> IntervalVarExpr:
        """Get the end time of interval variables.

        Returns
        -------
        IntervalVarExpr
            Expression representing the end time of each interval

        Examples
        --------
        >>> xmodel.add_constrs(df, end_constraint=var.task_iv.end_of() <= 100)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(end_of, return_dtype=pl.Object))

    def length_of(self) -> IntervalVarExpr:
        """Get the length of interval variables.

        Length is the actual work time (may differ from size/duration).

        Returns
        -------
        IntervalVarExpr
            Expression representing the length of each interval

        Examples
        --------
        >>> xmodel.add_constrs(df, length_constraint=var.task_iv.length_of() == 5)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(length_of, return_dtype=pl.Object))

    def size_of(self) -> IntervalVarExpr:
        """Get the size (duration) of interval variables.

        Size is the duration from start to end.

        Returns
        -------
        IntervalVarExpr
            Expression representing the size of each interval

        Examples
        --------
        >>> xmodel.add_constrs(df, size_constraint=var.task_iv.size_of() >= pl.col("min_duration"))  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(size_of, return_dtype=pl.Object))

    def presence_of(self) -> IntervalVarExpr:
        """Get the presence status of optional interval variables.

        Returns
        -------
        IntervalVarExpr
            Expression representing whether each interval is present (scheduled)

        Examples
        --------
        >>> xmodel.add_constrs(df, presence_constraint=var.task_iv.presence_of() == 1)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(presence_of, return_dtype=pl.Object))

    # Mathematical operations - Unary
    def abs(self) -> IntervalVarExpr:
        """Absolute value of the expression.

        Returns
        -------
        IntervalVarExpr
            Expression representing the absolute value

        Examples
        --------
        >>> xmodel.add_constrs(df, abs_constraint=var.task_iv.start_of().abs() <= 100)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(cp_abs, return_dtype=pl.Object))

    def square(self) -> IntervalVarExpr:
        """Square of the expression.

        Returns
        -------
        IntervalVarExpr
            Expression representing the square

        Examples
        --------
        >>> xmodel.add_constrs(df, square_constraint=var.task_iv.start_of().square() <= 100)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(square, return_dtype=pl.Object))

    def cp_log(self) -> IntervalVarExpr:
        """Natural logarithm of the expression (CP version).

        Returns
        -------
        IntervalVarExpr
            Expression representing the natural logarithm

        Examples
        --------
        >>> xmodel.add_constrs(df, log_constraint=var.task_iv.start_of().cp_log() >= 1)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(log, return_dtype=pl.Object))

    def exponent(self) -> IntervalVarExpr:
        """Exponential (e^x) of the expression.

        Returns
        -------
        IntervalVarExpr
            Expression representing e^x

        Examples
        --------
        >>> xmodel.add_constrs(df, exp_constraint=var.task_iv.start_of().exponent() <= 1000)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(exponent, return_dtype=pl.Object))

    def ceil(self) -> IntervalVarExpr:
        """Ceiling of the expression.

        Returns
        -------
        IntervalVarExpr
            Expression representing the ceiling

        Examples
        --------
        >>> xmodel.add_constrs(df, ceil_constraint=var.task_iv.start_of().ceil() == pl.col("rounded_start"))  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(ceil, return_dtype=pl.Object))

    def floor(self) -> IntervalVarExpr:
        """Floor of the expression.

        Returns
        -------
        IntervalVarExpr
            Expression representing the floor

        Examples
        --------
        >>> xmodel.add_constrs(df, floor_constraint=var.task_iv.start_of().floor() >= 0)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(floor, return_dtype=pl.Object))

    def cp_round(self) -> IntervalVarExpr:
        """Round to nearest integer (CP version).

        Returns
        -------
        IntervalVarExpr
            Expression representing the rounded value

        Examples
        --------
        >>> xmodel.add_constrs(df, round_constraint=var.task_iv.start_of().cp_round() == pl.col("start"))  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(cp_round, return_dtype=pl.Object))

    def trunc(self) -> IntervalVarExpr:
        """Truncate to integer (remove fractional part).

        Returns
        -------
        IntervalVarExpr
            Expression representing the truncated value

        Examples
        --------
        >>> xmodel.add_constrs(df, trunc_constraint=var.task_iv.start_of().trunc() >= 0)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(trunc, return_dtype=pl.Object))

    def sgn(self) -> IntervalVarExpr:
        """Sign of the expression (-1, 0, or 1).

        Returns
        -------
        IntervalVarExpr
            Expression representing the sign

        Examples
        --------
        >>> xmodel.add_constrs(df, sign_constraint=var.task_iv.start_of().sgn() >= 0)  # doctest: +SKIP

        """
        return IntervalVarExpr(self.map_elements(sgn, return_dtype=pl.Object))

    # Mathematical operations - Binary
    def power(self, exponent: VarExpr | pl.Expr | float) -> VarExpr:
        """Raise expression to a power.

        Parameters
        ----------
        exponent : VarExpr | pl.Expr | float
            The exponent

        Returns
        -------
        IntervalVarExpr
            Expression representing the power

        Examples
        --------
        >>> xmodel.add_constrs(df, power_constraint=var.task_iv.start_of().power(2) <= 100)  # doctest: +SKIP

        """
        return self._create_binary_operation(exponent, power)

    def minimum(self, other: VarExpr | pl.Expr | float) -> VarExpr:
        """Element-wise minimum of this expression and another value.

        Parameters
        ----------
        other : VarExpr | pl.Expr | float
            The other value to compare

        Returns
        -------
        IntervalVarExpr
            Expression representing the element-wise minimum

        Examples
        --------
        >>> xmodel.add_constrs(df, min_constraint=var.task_iv.start_of().minimum(10) >= 0)  # doctest: +SKIP

        """
        return self._create_binary_operation(other, cp_min)

    def maximum(self, other: VarExpr | pl.Expr | float) -> VarExpr:
        """Element-wise maximum of this expression and another value.

        Parameters
        ----------
        other : VarExpr | pl.Expr | float
            The other value to compare

        Returns
        -------
        VarExpr
            Expression representing the element-wise maximum

        Examples
        --------
        >>> xmodel.add_constrs(df, max_constraint=var.task_iv.start_of().maximum(0) <= 100)  # doctest: +SKIP

        """
        return self._create_binary_operation(other, cp_max)

    # Aggregation operations
    def sum(self) -> VarExpr:
        """Sum all values in the expression.

        Returns
        -------
        VarExpr
            Expression representing the sum (single value)

        Examples
        --------
        >>> xmodel.add_constrs(df, sum_constraint=var.task_iv.start_of().sum() <= 1000)  # doctest: +SKIP

        """

        def func(series: pl.Series) -> pl.Series:
            # Sum aggregates all values into a single result
            total = cp_sum(series.to_list())
            return pl.Series([total], dtype=pl.Object)

        return VarExpr(self.map_batches(func, return_dtype=pl.Object))


class _ProxyCplexCPVarExpr:
    """Proxy class for creating CPLEX CP interval variable expressions."""

    def __call__(self, name: str, /, *more_names: str) -> IntervalVarExpr:
        """Create an interval variable expression from column name(s).

        Parameters
        ----------
        name : str
            Primary column name
        *more_names : str
            Additional column names

        Returns
        -------
        IntervalVarExpr
            Expression wrapping the specified columns

        """
        return IntervalVarExpr(pl.col(name, *more_names))

    def __getattr__(self, name: str) -> IntervalVarExpr:
        """Create an interval variable expression via attribute access.

        Parameters
        ----------
        name : str
            Column name

        Returns
        -------
        IntervalVarExpr
            Expression wrapping the specified column

        """
        return IntervalVarExpr(pl.col(name))



"""CPLEX CP backend for constraint programming problems."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

import polars as pl
from docplex.cp.model import (
    CpoExpr,
    CpoModel,
    CpoVariable,
    binary_var_list,
    float_var_list,
    integer_var_list,
    interval_var,
)

from xplor._utils import parse_into_expr, series_to_df
from xplor.cplex_cp.var import IntervalVarExpr, _ProxyCplexCPVarExpr
from xplor.model import XplorModel
from xplor.types import cast_to_dtypes

if TYPE_CHECKING:
    from docplex.cp.expression import CpoIntervalVar
    from docplex.cp.solution import CpoSolveResult


class XplorCplexCP(XplorModel[CpoModel, CpoVariable, CpoExpr]):
    """Xplor wrapper for CPLEX CP (Constraint Programming) solver.

    This class provides a Polars-based interface for constraint programming problems
    like scheduling (RCPSP, job shop) using CPLEX CP Optimizer via docplex.cp.

    Unlike the mathematical programming backends (XplorCplex, XplorGurobi), this
    uses interval variables and CP-specific constraints.

    Attributes
    ----------
    model : docplex.cp.model.CpoModel
        The underlying CPLEX CP model

    Examples
    --------
    >>> import polars as pl  # doctest: +SKIP
    >>> from xplor.cplex_cp import XplorCplexCP  # doctest: +SKIP
    >>> xmodel = XplorCplexCP()  # doctest: +SKIP
    >>> # Create tasks DataFrame
    >>> df = pl.DataFrame({  # doctest: +SKIP
    ...     "task": ["T1", "T2", "T3"],
    ...     "duration": [3, 5, 4]
    ... })
    >>> # Add interval variables
    >>> df = df.with_columns(xmodel.add_interval_vars("iv", duration="duration"))  # doctest: +SKIP
    >>> # Add precedence constraint: T1 must finish before T2 starts
    >>> xmodel.add_constr(  # doctest: +SKIP
    ...     xmodel.var.iv.filter(pl.col("task") == "T1")
    ...         .end_before_start(xmodel.var.iv.filter(pl.col("task") == "T2"))
    ... )

    """

    def __init__(self, model: CpoModel | None = None) -> None:
        """Initialize the XplorCplexCP wrapper.

        Parameters
        ----------
        model : docplex.cp.model.CpoModel | None, default None
            Optional pre-existing CPLEX CP model. If None, creates a new model.

        """
        self.solution: CpoSolveResult | None = None
        super().__init__(CpoModel() if model is None else model)

    @cached_property
    def var(self) -> _ProxyCplexCPVarExpr:
        """Entry point for creating interval variable expressions.

        Similar to polars.col(), allows referencing interval variables created
        via add_interval_vars() in constraint expressions.

        Returns
        -------
        _ProxyCplexCPVarExpr
            Proxy object for creating interval variable expressions

        Examples
        --------
        >>> xmodel.var("task_iv")  # doctest: +SKIP
        >>> xmodel.var.task_iv  # doctest: +SKIP

        """
        return _ProxyCplexCPVarExpr()

    def read_values(self, name: pl.Expr) -> pl.Expr:
        """Read the value of an optimization variable from the solution.

        For interval variables, returns a struct with start, end, length, and present fields.
        For regular variables, returns the variable value.

        Parameters
        ----------
        name : pl.Expr
            Expression to evaluate.

        Returns
        -------
        pl.Expr
            Values of the variable expression. For interval variables, returns a struct
            with fields: start (Int64), end (Int64), length (Int64), present (Boolean).

        Examples
        --------
        >>> xmodel: XplorModel
        >>> # For regular variables
        >>> df_with_solution = df.with_columns(xmodel.read_values(pl.col("x")))  # doctest: +SKIP
        >>> # For interval variables - returns struct with start/end/length/present
        >>> df = df.with_columns(xmodel.read_values(pl.col("task_iv")))  # doctest: +SKIP
        >>> df = df.with_columns(  # doctest: +SKIP
        ...     start=pl.col("task_iv").struct.field("start"),
        ...     end=pl.col("task_iv").struct.field("end"),
        ... )

        """

        def _extract_interval(v: Any) -> dict[str, Any]:
            """Extract interval variable solution as struct."""
            if v is None or self.solution is None:
                return {"start": None, "end": None, "length": None, "present": None}
            sol = self.solution.get_var_solution(v)
            return {
                "start": sol.start,
                "end": sol.end,
                "length": sol.length,
                "present": sol.is_present(),
            }

        def _extract_regular(v: Any) -> float | None:
            """Extract regular variable solution value."""
            if v is None:
                return None
            if hasattr(v, "solution_value"):
                return v.solution_value
            return float(v)

        def _batch_extract(series: pl.Series) -> pl.Series:
            """Determine variable type and extract appropriately."""
            if len(series) == 0:
                return series

            # Check if this is an interval variable column (internal marker)
            if self.var_types.get(series.name) == "_INTERVAL_CP":
                # Return struct with start/end/length/present
                data = [_extract_interval(v) for v in series]
                return pl.Series(
                    series.name,
                    data,
                    dtype=pl.Struct(
                        {
                            "start": pl.Int64,
                            "end": pl.Int64,
                            "length": pl.Int64,
                            "present": pl.Boolean,
                        }
                    ),
                )
            else:
                # Regular variable - return scalar value
                return cast_to_dtypes(
                    pl.Series([_extract_regular(v) for v in series]),
                    self.var_types.get(series.name, "CONTINUOUS"),
                )

        return name.map_batches(_batch_extract)

    def _add_continuous_vars(
        self,
        names: list[str],
        lb: list[float] | None,
        ub: list[float] | None,
    ) -> list[CpoVariable]:
        return float_var_list(len(names), lb, ub, names)

    def _add_integer_vars(
        self,
        names: list[str],
        lb: list[float] | None,
        ub: list[float] | None,
    ) -> list[CpoVariable]:
        return integer_var_list(len(names), lb, ub, names)

    def _add_binary_vars(
        self,
        names: list[str],
    ) -> list[CpoVariable]:
        return binary_var_list(len(names), name=names)

    def add_interval_vars(
        self,
        name: str,
        *,
        start: float | tuple[int, int] | pl.Expr | None = None,
        end: float | tuple[int, int] | pl.Expr | None = None,
        duration: float | tuple[int, int] | pl.Expr | None = None,
        length: float | tuple[int, int] | pl.Expr | None = None,
        optional: bool | pl.Expr = False,
    ) -> pl.Expr:
        """Create interval variables for each row in the DataFrame.

        Interval variables represent tasks/activities with start time, end time,
        and duration. They can be optional (present/absent).

        Parameters
        ----------
        name : str
            Name for the interval variable column
        start : int | float | tuple[int, int] | pl.Expr | None, default None
            Start time or (min, max) bounds. If None, unbounded.
        end : int | float | tuple[int, int] | pl.Expr | None, default None
            End time or (min, max) bounds. If None, unbounded.
        duration : int | float | tuple[int, int] | pl.Expr | None, default None
            Duration or (min, max) bounds. If None, computed from start/end.
        length : int | float | tuple[int, int] | pl.Expr | None, default None
            Length (actual work time, may differ from duration). If None, equals duration.
        optional : bool | pl.Expr, default False
            Whether intervals can be absent (not scheduled)

        Returns
        -------
        pl.Expr
            Polars expression that creates interval variables when materialized

        Examples
        --------
        >>> df = df.with_columns(  # doctest: +SKIP
        ...     xmodel.add_interval_vars("task", duration=pl.col("task_duration"))
        ... )

        """
        return pl.map_batches(
            [
                parse_into_expr(start).alias("start"),  # ty:ignore[invalid-argument-type]
                parse_into_expr(end).alias("end"),  # ty:ignore[invalid-argument-type]
                parse_into_expr(duration).alias("duration"),  # ty:ignore[invalid-argument-type]
                parse_into_expr(length).alias("length"),  # ty:ignore[invalid-argument-type]
                parse_into_expr(optional).alias("optional"),
                pl.format(f"{name}[{{}}]", pl.row_index()).alias("var_name"),
            ],
            lambda s: self._add_interval_vars_wrapper(series_to_df(s), name),
            return_dtype=pl.Object,
        ).alias(name)

    def _add_interval_vars_wrapper(self, df: pl.DataFrame, name: str) -> pl.Series:
        """Create interval variables from a DataFrame of parameters.

        Parameters
        ----------
        df : pl.DataFrame
            DataFrame with columns: start, end, duration, length, optional, var_name
        name : str
            Name for the interval variable series

        Returns
        -------
        pl.Series
            Series of interval variables

        """
        # Create interval variables
        intervals = [
            interval_var(
                start=s,
                end=e,
                size=d,
                length=len_val,
                optional=o,
                name=n,
            )
            for s, e, d, len_val, o, n in df.with_columns(pl.col("optional").fill_null(False))
            .select("start", "end", "duration", "length", "optional", "var_name")
            .rows()
        ]

        # Store interval variables
        result = pl.Series(name, intervals, dtype=pl.Object)
        self.vars[name] = result
        # Use internal marker for interval variables (not part of public VariableType)
        self.var_types[name] = "_INTERVAL_CP"  # type: ignore
        return result

    def _add_constr(self, tmp_constr: Any, name: str) -> None:
        self.model.add(tmp_constr)

    def _materialize_interval_expr(self, expr: IntervalVarExpr | pl.Expr) -> list[CpoIntervalVar]:
        """Convert interval expression to list of actual interval variables.

        Parameters
        ----------
        expr : IntervalVarExpr | pl.Expr
            The expression to materialize

        Returns
        -------
        list[CpoIntervalVar]
            List of CPLEX CP interval variables

        """
        if isinstance(expr, IntervalVarExpr):
            # Get the column name from the expression
            # This is a simplified approach - may need refinement
            col_name = str(expr._name).replace("col(", "").replace(")", "").strip('"').strip("'")
            if col_name in self.vars:
                return self.vars[col_name].to_list()
            msg = f"Interval variable column '{col_name}' not found"
            raise ValueError(msg)
        elif isinstance(expr, pl.Expr):
            # Try to extract column name
            col_name = str(expr).replace("col(", "").replace(")", "").strip('"').strip("'")
            if col_name in self.vars:
                return self.vars[col_name].to_list()
            msg = f"Interval variable column '{col_name}' not found"
            raise ValueError(msg)
        else:
            msg = f"Invalid expression type: {type(expr)}"
            raise TypeError(msg)

    def minimize_makespan(self, intervals: pl.Expr | str) -> None:
        """Set objective to minimize makespan (maximum end time).

        Parameters
        ----------
        intervals : pl.Expr | str
            The interval variables column name or expression

        Examples
        --------
        >>> xmodel.minimize_makespan("task_iv")  # doctest: +SKIP

        """
        if isinstance(intervals, str):
            intervals = pl.col(intervals)

        ivars = self._materialize_interval_expr(intervals)
        # Minimize the maximum end time
        self.model.add(self.model.minimize(self.model.max([self.model.end_of(iv) for iv in ivars])))  # ty:ignore[unresolved-attribute]

    def optimize(self, **kwargs: Any) -> None:
        """Solve the CP model.

        Parameters
        ----------
        **kwargs : Any
            Additional parameters passed to model.solve()
            Common parameters:
            - TimeLimit: Maximum solve time in seconds
            - LogVerbosity: Verbosity level for solver output

        """
        self.solution = self.model.solve(**kwargs)

    def get_objective_value(self) -> float:
        """Return the objective value from the solved model.

        Returns
        -------
        float
            The objective value

        Raises
        ------
        ValueError
            If the model has not been solved

        """
        if self.solution is None:
            msg = "Model has not been optimized"
            raise ValueError(msg)
        return self.solution.get_objective_values()[0]



"""CPLEX CP backend for constraint programming with Polars."""

from xplor.cplex_cp.model import XplorCplexCP
from xplor.cplex_cp.var import _ProxyCplexCPVarExpr

var = _ProxyCplexCPVarExpr()
__all__ = ["XplorCplexCP", "var"]
